#!/usr/bin/env python3
import rclpy
from rclpy.node import Node
from std_msgs.msg import String
import numpy as np
import json  # To handle data serialization

def generate_path(H, points=101):
    rho = np.random.rand(H) * np.logspace(-0.5, -2.5, H)
    phi = np.random.rand(H) * 2 * np.pi
    t = np.linspace(0, 2 * np.pi, points)
    r = np.ones_like(t)
    for h in range(H):
        r += rho[h] * np.sin((h + 1) * t + phi[h])
    x = r * np.cos(t)
    y = r * np.sin(t)
    return x, y, t

def calculate_local_points(index, cones_x, cones_y, x, y, tangent_vectors_x, tangent_vectors_y):
    local_points = []
    car_pos = np.array([x[index], y[index]])
    car_dir = np.array([tangent_vectors_x[index], tangent_vectors_y[index]])
    R = np.array([[car_dir[0], -car_dir[1]],
                  [car_dir[1], car_dir[0]]])
    for cone_x, cone_y in zip(cones_x, cones_y):
        cone_pos = np.array([cone_x, cone_y])
        translated_point = cone_pos - car_pos
        local_point = R.T @ translated_point
        local_points.append(local_point)
    return local_points

def generate_cones(x, y, n, buffer):
    cones_x, cones_y = [], []
    for _ in range(n):
        index = np.random.randint(len(x))
        angle = 2 * np.pi * np.random.rand()
        radius = buffer + np.random.rand() * 0.5  # 0.5 to 1 units away from the buffer distance
        cx = x[index] + radius * np.cos(angle)
        cy = y[index] + radius * np.sin(angle)
        cones_x.append(cx)
        cones_y.append(cy)
    return np.array(cones_x), np.array(cones_y)

def visible_cones(x, y, cones_x, cones_y, index, tangent_vectors_x, tangent_vectors_y):
    visible_indices = []
    car_pos = np.array([x[index], y[index]])
    car_dir = np.array([tangent_vectors_x[index], tangent_vectors_y[index]])
    for i, (cx, cy) in enumerate(zip(cones_x, cones_y)):
        cone_vec = np.array([cx, cy]) - car_pos
        cone_angle = np.arccos(np.clip(np.dot(car_dir, cone_vec) / (np.linalg.norm(car_dir) * np.linalg.norm(cone_vec)), -1.0, 1.0))
        if cone_angle <= np.radians(60):  # 60 degrees to either side
            visible_indices.append(i)
    return visible_indices

class ConeVisibilityNode(Node):
    def __init__(self):
        super().__init__('cone_visibility_node')
        self.publisher_ = self.create_publisher(String, 'visible_cones', 10)
        timer_period = 0.02  # 50 Hz
        self.timer = self.create_timer(timer_period, self.timer_callback)
        self.x, self.y, self.t = generate_path(H=10)
        self.dx_dt = np.gradient(self.x, self.t)
        self.dy_dt = np.gradient(self.y, self.t)
        self.tangent_magnitudes = np.sqrt(self.dx_dt ** 2 + self.dy_dt ** 2)
        self.tangent_vectors_x = self.dx_dt / self.tangent_magnitudes
        self.tangent_vectors_y = self.dy_dt / self.tangent_magnitudes
        self.cones_x, self.cones_y = generate_cones(self.x, self.y, n=15, buffer=0.3)
        self.index = 70  # Update dynamically as needed

    def timer_callback(self):
        visible_indices = visible_cones(self.x, self.y, self.cones_x, self.cones_y, self.index, self.tangent_vectors_x, self.tangent_vectors_y)
        visible_cones_x = self.cones_x[visible_indices]
        visible_cones_y = self.cones_y[visible_indices]
        local_points = calculate_local_points(self.index, visible_cones_x, visible_cones_y, self.x, self.y, self.tangent_vectors_x, self.tangent_vectors_y)
        
        data = {
            "visible_cones": [
                {"id": idx + 1, "global": (self.cones_x[idx], self.cones_y[idx]), "local": (lp[0], lp[1])}
                for idx, lp in zip(visible_indices, local_points)
            ]
        }
        json_data = json.dumps(data)
        self.publisher_.publish(String(data=json_data))

def main(args=None):
    rclpy.init(args=args)
    cone_visibility_node = ConeVisibilityNode()
    rclpy.spin(cone_visibility_node)
    cone_visibility_node.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
